#!/usr/bin/env python
#
# suppress.py - Notification suppression.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides a couple of functions which may be used as context
managers for suppressing notification of changes to :class:`.HasProperties`
property values.

.. autosummary::
   :nosignatures:

   suppress
   suppressAll
   skip
"""


import contextlib


@contextlib.contextmanager
def suppress(hasProps, propName, notify=False):
    """Suppress notification for the given property on the given
    :class:`.HasProperties` instance.

    This function saves the current notification state of the
    property, disables notification, yields to the calling code,
    and then restores the notification state.


    :arg hasProps: ``HasProperties`` instance.

    :arg propName: Property to suppress notifications for.

    :arg notify:   If ``True``, a notification will be triggered
                   on ``propName`` via :meth:`.HasProperties.propNotify`,
                   exit. Defaults to ``False``.


    This function is intended to be used as follows::

        with suppress(hasProps1, 'prop1'):
            # Do stuff which might cause unwanted
            # property value notifications to occur
    """

    state = hasProps.getNotificationState(propName)
    hasProps.disableNotification(propName)

    try:
        yield

    finally:
        hasProps.setNotificationState(propName, state)

    if notify:
        hasProps.propNotify(propName)


@contextlib.contextmanager
def suppressAll(hasProps):
    """Suppress notification for all properties on the given
    :class:`.HasProperties` instance.

    :arg hasProps: The ``HasProperties`` instance to suppress.


    This function is intended to be used as follows::

        with suppressAll(hasProps):
            # Do stuff which might cause unwanted
            # property value notifications to occur


    .. note:: After this function has completed. notification
              will be enabled for all properties of ``hasProps``.
              This is in contrast to the :func:`suppress` function,
              which restores the previous notification state of
              the property.
    """

    hasProps.disableAllNotification()

    try:
        yield
    finally:
        hasProps.enableAllNotification()


@contextlib.contextmanager
def skip(hasProps, propName, listenerName):
    """Skip the listener with the specified listener name, if/when
    changes to the specified property occur.

    :arg hasProps:     A :class:`.HasProperties` instance.
    :arg propName:     Name of a property on ``hasProps``.
    :arg listenerName: Name of the listener to skip.
    """

    state = hasProps.getListenerState(propName, listenerName)

    hasProps.disableListener(propName, listenerName)

    try:
        yield

    finally:
        hasProps.setListenerState(propName, listenerName, state)



#!/usr/bin/env python
#
# properties_value.py - Definitions of the PropertyValue and
#                       PropertyValueList classes.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""Definitions of the :class:`PropertyValue` and :class:`PropertyValueList`
classes.

 .. autosummary::
    :nosignatures:

    PropertyValue
    PropertyValueList


``PropertyValue`` and ``PropertyValueList`` instances are intended to be
created and managed by :class:`.PropertyBase` and :class:`.ListPropertyBase`
instances respectively, and are used to encapsulate attribute values of
:class:`.HasProperties` instances.


These class definitions are really a part of the :mod:`.properties` module -
they are separated to keep file sizes down.  However, the
:class:`.PropertyValue` class definitions have no dependence upon the
:class:`.PropertyBase` or :class:`.HasProperties` definitions.
"""


import uuid
import logging
import weakref

from collections import OrderedDict

from . import callqueue
from . import bindable

import fsl.utils.weakfuncref as weakfuncref


log = logging.getLogger(__name__)


class Listener(object):
    """The ``Listener`` class is used by :class:`PropertyValue` instances to
    manage their listeners - see :meth:`PropertyValue.addListener`.
    """
    def __init__(self, propVal, name, function, enabled, immediate):
        """Create a ``Listener``.

        :arg propVal:   The ``PropertyValue`` that owns this ``Listener``.
        :arg name:      The listener name.
        :arg function:  The callback function.
        :arg enabled:   Whether the listener is enabled/disabled.
        :arg immediate: Whether the listener is to be called immediately, or
                        via the :attr:`PropertyValue.queue`.
        """

        self.propVal   = weakref.ref(propVal)
        self.name      = name
        self.function  = function
        self.enabled   = enabled
        self.immediate = immediate


    def makeQueueName(self):
        """Returns a more descriptive name for this ``Listener``, which
        is used as its name when passed to the :class:`.CallQueue`.
        """

        ctxName = self.propVal()._context().__class__.__name__
        pvName  = self.propVal()._name

        return '{} ({}.{})'.format(self.name, ctxName, pvName)


class PropertyValue(object):
    """An object which encapsulates a value of some sort.

    The value may be subjected to casting and validation rules, and listeners
    may be registered for notification of value and validity changes.

    Notification of value and attribute listeners is performed by the
    :mod:`.bindable` module - see the :func:`.bindable.syncAndNotify` and
    :func:`.bindable.syncAndNotifyAtts` functions.
    """


    queue = callqueue.CallQueue(skipDuplicates=True)
    """A :class:`.CallQueue` instance which is shared by all
    :class:`PropertyValue` instances, and used for notifying listeners
    of value and attribute changes.

    A queue is used for notification so that listeners are notified in
    the order that values were changed.
    """


    def __init__(self,
                 context,
                 name=None,
                 value=None,
                 castFunc=None,
                 validateFunc=None,
                 equalityFunc=None,
                 preNotifyFunc=None,
                 postNotifyFunc=None,
                 allowInvalid=True,
                 parent=None,
                 **attributes):
        """Create a ``PropertyValue`` object.

        :param context:        An object which is passed as the first argument
                               to the ``validateFunc``, ``preNotifyFunc``,
                               ``postNotifyFunc``, and any registered
                               listeners. Can technically be anything, but will
                               nearly always be a :class:`.HasProperties`
                               instance.

        :param name:           Value name - if not provided, a default, unique
                               name is created.

        :param value:          Initial value.

        :param castFunc:       Function which performs type casting or data
                               conversion. Must accept three parameters - the
                               context, a dictionary containing the attributes
                               of this object, and the value to cast. Must
                               return that value, cast/converted appropriately.

        :param validateFunc:   Function which accepts three parameters - the
                               context, a dictionary containing the attributes
                               of this object, and a value. This function
                               should test the provided value, and raise a
                               :exc:`ValueError` if it is invalid.

        :param equalityFunc:   Function which accepts two values, and should
                               return ``True`` if they are equal, ``False``
                               otherwise. If not provided, the python equailty
                               operator (i.e. ``==``) is used.

        :param preNotifyFunc:  Function to be called whenever the property
                               value changes, but before any registered
                               listeners are called. See the
                               :meth:`addListener` method for details of the
                               parameters this function must accept.

        :param postNotifyFunc: Function to be called whenever the property
                               value changes, but after any registered
                               listeners are called. Must accept the same
                               parameters as the ``preNotifyFunc``.

        :param allowInvalid:   If ``False``, any attempt to set the value to
                               something invalid will result in a
                               :exc:`ValueError`. Note that this does not
                               guarantee that the property will never have an
                               invalid value, as the definition of 'valid'
                               depends on external factors (i.e. the
                               ``validateFunc``).  Therefore, the validity of
                               a value may change, even if the value itself
                               has not changed.

        :param parent:         If this PV instance is a member of a
                               :class:`PropertyValueList` instance, the latter
                               sets itself as the parent of this PV. Whenever
                               the value of this PV changes, the
                               :meth:`PropertyValueList._listPVChanged` method
                               is called.

        :param attributes:     Any key-value pairs which are to be associated
                               with this :class:`PropertyValue` object, and
                               passed to the ``castFunc`` and ``validateFunc``
                               functions. Attributes are not used by the
                               ``PropertyValue`` or ``PropertyValueList``
                               classes, however they are used by the
                               :class:`.PropertyBase` and
                               :class:`.ListPropertyBase` classes to store
                               per-instance property constraints. Listeners
                               may register to be notified when attribute
                               values change.
        """

        if name     is     None: name  = 'PropertyValue_{}'.format(id(self))
        if castFunc is not None: value = castFunc(context, attributes, value)
        if equalityFunc is None: equalityFunc = lambda a, b: a == b

        self._context                 = weakref.ref(context)
        self._validate                = validateFunc
        self._name                    = name
        self._equalityFunc            = equalityFunc
        self._castFunc                = castFunc
        self._allowInvalid            = allowInvalid
        self._attributes              = attributes.copy()
        self._changeListeners         = OrderedDict()
        self._attributeListeners      = OrderedDict()

        self.__value                  = value
        self.__valid                  = False
        self.__lastValue              = None
        self.__lastValid              = False
        self.__notification           = True

        self._preNotifyListener  = Listener(self,
                                            'prenotify',
                                            preNotifyFunc,
                                            True,
                                            True)
        self._postNotifyListener = Listener(self,
                                            'postnotify',
                                            postNotifyFunc,
                                            True,
                                            False)

        if parent is not None: self.__parent = weakref.ref(parent)
        else:                  self.__parent = None

        if not allowInvalid and validateFunc is not None:
            validateFunc(context, self._attributes, value)


    def __repr__(self):
        """Returns a string representation of this PropertyValue object."""
        return 'PV({})'.format(self.__value)


    def __str__(self):
        """Returns a string representation of this PropertyValue object."""
        return self.__repr__()


    def __hash__(self):
        return id(self)


    def __eq__(self, other):
        """Returns ``True`` if the given object has the same value as this
        instance. Returns ``False`` otherwise.
        """
        if isinstance(other, PropertyValue):
            other = other.get()
        return self._equalityFunc(self.get(), other)


    def __ne__(self, other):
        """Returns ``True`` if the given object has a different value to
        this instance, ``False`` otherwise.
        """
        return not self.__eq__(other)


    def __saltListenerName(self, name):
        """Adds a constant string to the given listener name.

        This is done for debug output, so we can better differentiate between
        listeners with the same name registered on different PV objects.
        """
        return 'PropertyValue_{}_{}'.format(self._name, name)


    def __unsaltListenerName(self, name):
        """Removes a constant string from the given listener name,
        which is assumed to have been generated by the
        :meth:`__saltListenerName` method.
        """

        salt = 'PropertyValue_{}_'.format(self._name)

        return name[len(salt):]


    def getParent(self):
        """If this ``PropertyValue`` is an item in a :class:`PropertyValueList`,
        this method returns a reference to the owning ``PropertyValueList``.
        Otherwise, this method returns ``None``.
        """
        if self.__parent is not None: return self.__parent()
        else:                         return None


    def allowInvalid(self, allow=None):
        """Query/set the allow invalid state of this value.

        If no arguments are passed, returns the current allow invalid state.
        Otherwise, sets the current allow invalid state. to the given argument.
        """
        if allow is None:
            return self._allowInvalid

        self._allowInvalid = bool(allow)


    def enableNotification(self, bound=False, att=False):
        """Enables notification of property value and attribute listeners for
        this ``PropertyValue`` object.

        :arg bound: If ``True``, notification is enabled on all other
                    ``PropertyValue`` instances that are bound to this one
                    (see the :mod:`.bindable` module). If ``False`` (the
                    default), notification is only enabled on this
                    ``PropertyValue``.

        :arg att:   If ``True``, notification is enabled on all attribute
                    listeners as well as property value listeners.
        """
        self.__notification = True

        if not bound:
            return

        bpvs = list(bindable.buildBPVList(self, 'boundPropVals')[0])

        if att:
            bpvs += list(bindable.buildBPVList(self, 'boundAttPropVals')[0])

        for bpv in bpvs:
            bpv.enableNotification()


    def disableNotification(self, bound=False, att=False):
        """Disables notification of property value and attribute listeners for
        this ``PropertyValue`` object. Notification can be re-enabled via
        the :meth:`enableNotification` or :meth:`setNotificationState` methods.


        :arg bound: If ``True``, notification is disabled on all other
                    ``PropertyValue`` instances that are bound to this one
                    (see the :mod:`.bindable` module). If ``False`` (the
                    default), notification is only disabled on this
                    ``PropertyValue``.

        :arg att:   If ``True``, notification is disabled on all attribute
                    listeners as well as property value listeners.
        """
        self.__notification = False

        if not bound:
            return

        bpvs = list(bindable.buildBPVList(self, 'boundPropVals')[0])

        if att:
            bpvs += list(bindable.buildBPVList(self, 'boundAttPropVals')[0])

        for bpv in bpvs:
            bpv.disableNotification()


    def getNotificationState(self):
        """Returns ``True`` if notification is currently enabled, ``False``
        otherwise.
        """
        return self.__notification


    def setNotificationState(self, value):
        """Sets the current notification state."""
        if value: self.enableNotification()
        else:     self.disableNotification()


    def addAttributeListener(self, name, listener, weak=True, immediate=False):
        """Adds an attribute listener for this ``PropertyValue``. The
        listener callback function must accept the following arguments:

          - ``context``:   The context associated with this ``PropertyValue``.

          - ``attribute``: The name of the attribute that changed.

          - ``value``:     The new attribute value.

          - ``name``:      The name of this ``PropertyValue`` instance.

        :param name:      A unique name for the listener. If a listener with
                          the specified name already exists, it will be
                          overwritten.

        :param listener:  The callback function.

        :param weak:      If ``True`` (the default), a weak reference to the
                          callback function is used.

        :param immediate: If ``False`` (the default), the listener is called
                          immediately; otherwise, it is called via the
                          :attr:`queue`.
        """
        log.debug('Adding attribute listener on {}.{} ({}): {}'.format(
            self._context().__class__.__name__, self._name, id(self), name))

        if weak:
            listener = weakfuncref.WeakFunctionRef(listener)

        name = self.__saltListenerName(name)
        self._attributeListeners[name] = Listener(self,
                                                  name,
                                                  listener,
                                                  True,
                                                  immediate)


    def disableAttributeListener(self, name):
        """Disables the attribute listener with the specified ``name``. """
        name = self.__saltListenerName(name)
        log.debug('Disabling attribute listener on {}: {}'.format(self._name,
                                                                  name))
        self._attributeListeners[name].enabled = False


    def enableAttributeListener(self, name):
        """Enables the attribute listener with the specified ``name``. """
        name = self.__saltListenerName(name)
        log.debug('Enabling attribute listener on {}: {}'.format(self._name,
                                                                 name))
        self._attributeListeners[name].enabled = True


    def removeAttributeListener(self, name):
        """Removes the attribute listener of the given name."""
        log.debug('Removing attribute listener on {}.{}: {}'.format(
            self._context().__class__.__name__, self._name, name))

        name     = self.__saltListenerName(name)
        listener = self._attributeListeners.pop(name, None)

        if listener is not None:

            cb = listener.function

            if isinstance(cb, weakfuncref.WeakFunctionRef):
                cb = cb.function()

            if cb is not None:
                PropertyValue.queue.dequeue(listener.makeQueueName())


    def getAttributes(self):
        """Returns a dictionary containing all the attributes of this
        ``PropertyValue`` object.
        """
        return self._attributes.copy()


    def setAttributes(self, atts):
        """Sets all the attributes of this ``PropertyValue`` object.
        from the given dictionary.
        """

        for name, value in atts.items():
            self.setAttribute(name, value)


    def getAttribute(self, name):
        """Returns the value of the named attribute."""
        return self._attributes[name]


    def setAttribute(self, name, value):
        """Sets the named attribute to the given value, and notifies any
        registered attribute listeners of the change.
        """
        oldVal = self._attributes.get(name, None)

        self._attributes[name] = value

        if oldVal == value: return

        log.debug('Attribute on {}.{} ({}) changed: {} = {}'.format(
            self._context().__class__.__name__,
            self._name,
            id(self),
            name,
            value))

        self.notifyAttributeListeners(name, value)

        self.revalidate()


    def prepareListeners(self, att, name=None, value=None):
        """Prepares a list of :class:`Listener` instances ready to be called,
        and a list of arguments to pass to them.

        :arg att:   If ``True``, attribute listeners are returned, otherwise
                    value listeners are returned.
        :arg name:  If ``att == True``, the attribute name.
        :arg value: If ``att == True``, the attribute value.
        """

        if not self.__notification:
            return [], []

        if att: lDict = self._attributeListeners
        else:   lDict = self._changeListeners

        allListeners = []

        for lName, listener in list(lDict.items()):

            if not listener.enabled:
                continue

            cb = listener.function

            if isinstance(cb, weakfuncref.WeakFunctionRef):
                cb = cb.function()

            # The owner of the referred function/method
            # has been GC'd - remove it
            if cb is None:

                log.debug('Removing dead listener {}'.format(lName))
                if att:
                    self.removeAttributeListener(
                        self.__unsaltListenerName(lName))
                else:
                    self.removeListener(
                        self.__unsaltListenerName(lName))
                continue

            allListeners.append(listener)

        # if we're preparing value listenres, add
        # the pre-notify and post-notify functions
        if not att:

            if self._preNotifyListener.function is not None and \
               self._preNotifyListener.enabled:
                allListeners = [self._preNotifyListener] + allListeners

            if self._postNotifyListener.function is not None and \
               self._postNotifyListener.enabled:
                allListeners = allListeners + [self._postNotifyListener]

        if att: args = self._context(), name, value, self._name
        else:   args = (self.get(), self.__valid, self._context(), self._name)

        return allListeners, args


    def notifyAttributeListeners(self, name, value):
        """Notifies all registered attribute listeners of an attribute
        changed - see the :func:`.bindable.syncAndNotifyAtts` function.
        """

        bindable.syncAndNotifyAtts(self, name, value)


    def addListener(self,
                    name,
                    callback,
                    overwrite=False,
                    weak=True,
                    immediate=False):
        """Adds a listener for this value.

        When the value changes, the listener callback function is called. The
        callback function must accept the following arguments:

          - ``value``:   The property value
          - ``valid``:   Whether the value is valid or invalid
          - ``context``: The context object passed to :meth:`__init__`.
          - ``name``:    The name of this ``PropertyValue`` instance.

        Listener names ``prenotify`` and ``postnotify`` are reserved - if
        either of these are passed in for the listener name, a :exc`ValueError`
        is raised.

        :param str name:  A unique name for this listener. If a listener with
                          the name already exists, a :exc`RuntimeError` will be
                          raised, or it will be overwritten, depending upon
                          the value of the ``overwrite`` argument.

        :param callback:  The callback function.

        :param overwrite: If ``True`` any previous listener with the same name
                          will be overwritten.

        :param weak:      If ``True`` (the default), a weak reference to the
                          callback function is retained, meaning that it
                          can be garbage-collected. If passing in a lambda or
                          inner function, you will probably want to set
                          ``weak`` to ``False``, in which case a strong
                          reference will be used.

        :param immediate: If ``False`` (the default), this listener will be
                          notified through the :class:`.CallQueue` - listeners
                          for all ``PropertyValue`` instances are queued, and
                          notified in turn. If ``True``, If ``True``, the
                          ``CallQueue`` will not be used, and this listener
                          will be notified as soon as this ``PropertyValue``
                          changes.
        """

        if name in ('prenotify', 'postnotify'):
            raise ValueError('Reserved listener name used: {}. '
                             'Use a different name.'.format(name))

        log.debug('Adding listener on {}.{}: {}'.format(
            self._context().__class__.__name__,
            self._name,
            name))

        fullName = self.__saltListenerName(name)
        prior    = self._changeListeners.get(fullName, None)

        if weak:
            callback = weakfuncref.WeakFunctionRef(callback)

        if (prior is not None) and (not overwrite):
            raise RuntimeError('Listener {} already exists'.format(name))

        elif prior is not None:
            prior.function  = callback
            prior.immediate = immediate

        else:
            self._changeListeners[fullName] = Listener(self,
                                                       fullName,
                                                       callback,
                                                       True,
                                                       immediate)


    def removeListener(self, name):
        """Removes the listener with the given name from this
        ``PropertyValue``.
        """

        # The typical stack trace of a call to this method is:
        #    someHasPropertiesObject.removeListener(...) (the original call)
        #      HasProperties.removeListener(...)
        #        PropertyBase.removeListener(...)
        #          this method
        # So to be a bit more informative, we'll examine the stack
        # and extract the (assumed) location of the original call
        if log.getEffectiveLevel() == logging.DEBUG:
            import inspect
            stack = inspect.stack()

            if len(stack) >= 4: frame = stack[ 3]
            else:               frame = stack[-1]

            srcMod  = '...{}'.format(frame[1][-20:])
            srcLine = frame[2]

            log.debug('Removing listener on {}.{}: {} ({}:{})'.format(
                self._context().__class__.__name__,
                self._name,
                name,
                srcMod,
                srcLine))

        name     = self.__saltListenerName(name)
        listener = self._changeListeners.pop(name, None)

        if listener is not None:

            # The bindable._allAllListeners does
            # funky things to the call queue,
            # so we mark this listener as disabled
            # just in case bindable tries to call
            # a removed listener.
            listener.enabled = False

            cb = listener.function

            if isinstance(cb, weakfuncref.WeakFunctionRef):
                cb = cb.function()

            if cb is not None:
                PropertyValue.queue.dequeue(listener.makeQueueName())


    def enableListener(self, name):
        """(Re-)Enables the listener with the specified ``name``."""
        name = self.__saltListenerName(name)
        log.debug('Enabling listener on {}: {}'.format(self._name, name))
        self._changeListeners[name].enabled = True


    def disableListener(self, name):
        """Disables the listener with the specified ``name``, but does not
        remove it from the list of listeners.
        """
        name = self.__saltListenerName(name)
        log.debug('Disabling listener on {}: {}'.format(self._name, name))
        self._changeListeners[name].enabled = False


    def getListenerState(self, name):
        """Returns ``True`` if the specified listener is currently enabled,
        ``False`` otherwise.

        An :exc:`AttributeError` is raised if a listener with the specified
        ``name`` does not exist.
        """

        fullName = self.__saltListenerName(name)
        listener = self._changeListeners.get(fullName, None)

        return listener.enabled


    def setListenerState(self, name, state):
        """Enables/disables the specified listener. """

        if state: self.enableListener(name)
        else:     self.disableListener(name)


    def hasListener(self, name):
        """Returns ``True`` if a listener with the given name is registered,
        ``False`` otherwise.
        """

        name = self.__saltListenerName(name)
        return name in self._changeListeners.keys()


    def setPreNotifyFunction(self, preNotifyFunc):
        """Sets the function to be called on value changes, before any
        registered listeners.
        """
        self._preNotifyListener.function = preNotifyFunc


    def setPostNotifyFunction(self, postNotifyFunc):
        """Sets the function to be called on value changes, after any
        registered listeners.
        """
        self._postNotifyListener.function = postNotifyFunc


    def getLast(self):
        """Returns the most recent property value before the current one."""
        return self.__lastValue


    def get(self):
        """Returns the current property value."""
        return self.__value


    def set(self, newValue):
        """Sets the property value.

        The property is validated and, if the property value or its validity
        has changed, any registered listeners are called through the
        :meth:`propNotify` method.  If ``allowInvalid`` was set to
        ``False``, and the new value is not valid, a :exc:`ValueError` is
        raised, and listeners are not notified.
        """

        # cast the value if necessary.
        # Allow any errors to be thrown
        if self._castFunc is not None:
            newValue = self._castFunc(self._context(),
                                      self._attributes,
                                      newValue)

        # Check to see if the new value is valid
        valid    = False
        validStr = None
        try:
            if self._validate is not None:
                self._validate(self._context(), self._attributes, newValue)
            valid = True

        except ValueError as e:

            # Oops, we don't allow invalid values.
            validStr = str(e)
            if not self._allowInvalid:
                import traceback
                log.debug('Attempt to set {}.{} to an invalid value ({}), '
                          'but allowInvalid is False ({})'.format(
                              self._context().__class__.__name__,
                              self._name,
                              newValue,
                              e), exc_info=True)
                traceback.print_stack()
                raise e

        self.__lastValue = self.__value
        self.__lastValid = self.__valid
        self.__value     = newValue
        self.__valid     = valid

        # If the value or its validity has not
        # changed, listeners are not notified
        changed = (self.__valid != self.__lastValid) or \
                  not self._equalityFunc(self.__value, self.__lastValue)

        if not changed: return

        log.debug('Value {}.{} changed: {} -> {} ({})'.format(
            self._context().__class__.__name__,
            self._name,
            self.__lastValue,
            self.__value,
            'valid' if valid else 'invalid - {}'.format(validStr)))

        # Notify any registered listeners.
        self.propNotify()


    def propNotify(self):
        """Notifies registered listeners - see the
        :func:`.bindable.syncAndNotify` function.
        """

        bindable.syncAndNotify(self)

        if not self.__notification:
            return

        # If this PV is a member of a PV list,
        # tell the list that this PV has
        # changed, so that it can notify its own
        # list-level listeners of the change
        if self.__parent is not None and self.__parent() is not None:
            self.__parent()._listPVChanged(self)


    def revalidate(self):
        """Revalidates the current property value, and re-notifies any
        registered listeners if the value validity has changed.
        """
        self.set(self.get())


    def isValid(self):
        """Returns ``True`` if the current property value is valid, ``False``
        otherwise.
        """
        try: self._validate(self._context(), self._attributes, self.get())
        except: return False
        return True


class PropertyValueList(PropertyValue):
    """A ``PropertyValueList`` is a :class:`PropertyValue` instance which
    stores other :class:`PropertyValue` instance in a list. Instances of
    this class are generally managed by a :class:`.ListPropertyBase` instance.

    When created, separate validation functions may be passed in for
    individual items, and for the list as a whole. Listeners may be registered
    on individual ``PropertyValue`` instances (accessible via the
    :meth:`getPropertyValueList` method), or on the entire list.

    The values contained in this ``PropertyValueList`` may be accessed
    through standard Python list operations, including slice-based access and
    assignment, :meth:`append`, :meth:`insert`, :meth:`extend`, :meth:`pop`,
    :meth:`index`, :meth:`count`, :meth:`move`, :meth:`insertAll`,
    :meth:`removeAll`, and :meth:`reorder` (these last few are non-standard).

    Because the values contained in this list are ``PropertyValue``
    instances themselves, some limitations are present on list modifying
    operations::

      class MyObj(props.HasProperties):
        mylist = props.List(default[1, 2, 3])

      myobj = MyObj()

    Simple list-slicing modifications work as expected::

      # the value after this will be [5, 2, 3]
      myobj.mylist[0]  = 5

      # the value after this will be [5, 6, 7]
      myobj.mylist[1:] = [6, 7]

    However, modifications which would change the length of the list are not
    supported::

      # This will result in an IndexError
      myobj.mylist[0:2] = [6, 7, 8]

    The exception to this rule concerns modifications which would replace
    every value in the list::

      # These assignments are equivalent
      myobj.mylist[:] = [1, 2, 3, 4, 5]
      myobj.mylist    = [1, 2, 3, 4, 5]

    While the simple list modifications described above will change the
    value(s) of the existing ``PropertyValue`` instances in the list,
    modifications which replace the entire list contents will result in
    existing ``PropertyValue`` instances being destroyed, and new ones
    being created. This is a very important point to remember if you have
    registered listeners on individual ``PropertyValue`` items.

    A listener registered on a ``PropertyValueList`` will be notified
    whenever the list is modified (e.g. additions, removals, reorderings), and
    whenever any individual value in the list changes. Alternately, listeners
    may be registered on the individual ``PropertyValue`` instances (which
    are accessible through the :meth:`getPropertyValueList` method) to be
    nofitied of changes to those values only.

    There are some interesting type-specific subclasses of the
    ``PropertyValueList``, which provide additional functionality:

      - The :class:`.PointValueList`, for :class:`.Point` properties.

      - The :class:`.BoundsValueList`, for :class:`.Bounds` properties.
    """

    def __init__(self,
                 context,
                 name=None,
                 values=None,
                 itemCastFunc=None,
                 itemEqualityFunc=None,
                 itemValidateFunc=None,
                 listValidateFunc=None,
                 itemAllowInvalid=True,
                 preNotifyFunc=None,
                 postNotifyFunc=None,
                 listAttributes=None,
                 itemAttributes=None):
        """Create a ``PropertyValueList``.

        :param context:          See :meth:`PropertyValue.__init__`.

        :param name:             See :meth:`PropertyValue.__init__`.

        :param values:           Initial list values.

        :param itemCastFunc:     Function which casts a single list item.

        :param itemEqualityFunc: Function which tests equality of two values.

        :param itemValidateFunc: Function which validates a single list item.

        :param listValidateFunc: Function which validates the list as a whole.

        :param itemAllowInvalid: Whether items are allowed to containg
                                 invalid values.

        :param preNotifyFunc:    See :meth:`PropertyValue.__init__`.

        :param postNotifyFunc:   See :meth:`PropertyValue.__init__`.

        :param listAttributes:   Attributes to be associated with this
                                 ``PropertyValueList``.

        :param itemAttributes:   Attributes to be associated with new
                                 ``PropertyValue`` items added to
                                 the list.
        """
        if name is None: name = 'PropertyValueList_{}'.format(id(self))

        if listAttributes is None: listAttributes = {}

        def itemEquals(a, b):
            if isinstance(a, PropertyValue): a = a.get()
            if isinstance(b, PropertyValue): b = b.get()

            if itemEqualityFunc is not None: return itemEqualityFunc(a, b)
            else:                            return a == b

        listValid = None
        if listValidateFunc is not None:
            def listValid(ctx, atts, value):
                value = list(value)
                for i, v in enumerate(value):
                    if isinstance(v, PropertyValue):
                        value[i] = v.get()
                return listValidateFunc(ctx, atts, value)

        # The list as a whole must be allowed to contain
        # invalid values because, if an individual
        # PropertyValue item value changes, there is no
        # nice way to propagate those changes on to other
        # (dependent) items without the list as a whole
        # being validated first, and errors being raised.
        PropertyValue.__init__(
            self,
            context,
            name=name,
            allowInvalid=True,
            validateFunc=listValid,
            preNotifyFunc=preNotifyFunc,
            postNotifyFunc=postNotifyFunc,
            **listAttributes)

        # These attributes are passed to the PropertyValue
        # constructor whenever a new item is added to the list
        self._itemCastFunc     = itemCastFunc
        self._itemValidateFunc = itemValidateFunc
        self._itemEqualityFunc = itemEquals
        self._itemAllowInvalid = itemAllowInvalid
        self._itemAttributes   = itemAttributes

        # The list of PropertyValue objects.
        if values is not None: values = [self.__newItem(v) for v in values]
        else:                  values = []

        PropertyValue.set(self, values)


    def __eq__(self, other):
        """Retuns ``True`` if the given object contains the same values as
        this instance, ``False`` otherwise.
        """

        if other is None:
            return False

        if len(self) != len(other):
            return False

        return all([self._itemEqualityFunc(ai, bi)
                    for ai, bi
                    in zip(self[:], other[:])])


    def getPropertyValueList(self):
        """Return (a copy of) the underlying property value list, allowing
        access to the ``PropertyValue`` instances which manage each list
        item.
        """
        return list(PropertyValue.get(self))


    def get(self):
        """Overrides :meth:`PropertyValue.get`. Returns this
        ``PropertyValueList`` object.
        """
        return self


    def set(self, newValues):
        """Overrides :meth:`PropertyValue.set`.

        Sets the values stored in this ``PropertyValueList``. If the
        length of the ``newValues`` argument does not match the current list
        length,  an :exc:`IndexError` is raised.
        """

        if self._itemCastFunc is not None:
            newValues = [self._itemCastFunc(
                self._context(),
                self._itemAttributes,
                v) for v in newValues]

        self[:] = newValues


    def __newItem(self, item):
        """Called whenever a new item is added to the list.  Encapsulate the
        given item in a ``PropertyValue`` instance.
        """

        if self._itemAttributes is None: itemAtts = {}
        else:                            itemAtts = self._itemAttributes

        propVal = PropertyValue(
            self._context(),
            name='{}_Item'.format(self._name),
            value=item,
            castFunc=self._itemCastFunc,
            allowInvalid=self._itemAllowInvalid,
            equalityFunc=self._itemEqualityFunc,
            validateFunc=self._itemValidateFunc,
            parent=self,
            **itemAtts)

        return propVal


    def getLast(self):
        """Overrides :meth:`PropertyValue.getLast`. Returns the most
        recent list values.
        """
        return [pv.get() for pv in PropertyValue.getLast(self)]


    def _listPVChanged(self, pv):
        """This function is called by list items when their value changes.
        List-level listeners are notified of the change. See the
        :meth:`PropertyValue.propNotify` method.
        """

        if self.getNotificationState():
            log.debug('List item {}.{} changed ({}) - nofiying '
                      'list-level listeners ({})'.format(
                          self._context().__class__.__name__,
                          self._name,
                          pv,
                          self[:]))
            self.propNotify()


    def __getitem__(self, key):
        vals = [pv.get() for pv in PropertyValue.get(self)]
        return vals.__getitem__(key)

    def __len__(     self):
        return self[:].__len__()
    def __repr__(    self):
        return self[:].__repr__()
    def __str__(     self):
        return self[:].__str__()
    def __iter__(    self):
        return self[:].__iter__()
    def __contains__(self, item):
        return self[:].__contains__(item)
    def index(       self, item):
        return self[:].index(item)
    def count(       self, item):
        return self[:].count(item)


    def insert(self, index, item):
        """Inserts the given item before the given index. """

        propVals = self.getPropertyValueList()
        propVals.insert(index, self.__newItem(item))
        PropertyValue.set(self, propVals)


    def insertAll(self, index, items):
        """Inserts all of the given items before the given index."""

        propVals = self.getPropertyValueList()
        propVals[index:index] = [self.__newItem(i) for i in items]
        PropertyValue.set(self, propVals)


    def append(self, item):
        """Appends the given item to the end of the list."""

        propVals = self.getPropertyValueList()
        propVals.append(self.__newItem(item))
        PropertyValue.set(self, propVals)


    def extend(self, iterable):
        """Appends all items in the given iterable to the end of the list."""

        propVals = self.getPropertyValueList()
        propVals.extend([self.__newItem(i) for i in iterable])
        PropertyValue.set(self, propVals)


    def pop(self, index=-1):
        """Remove and return the specified value in the list (default:
        last).
        """

        propVals      = self.getPropertyValueList()
        poppedPropVal = propVals.pop(index)
        PropertyValue.set(self, propVals)
        return poppedPropVal.get()


    def move(self, from_, to):
        """Move the item from 'from\_' to 'to'."""

        propVals = self.getPropertyValueList()
        propVals.insert(to, propVals.pop(from_))
        PropertyValue.set(self, propVals)


    def remove(self, value):
        """Remove the first item in the list with the specified value. """

        # delegates to __delitem__, defined below
        del self[self.index(value)]


    def removeAll(self, values):
        """Removes the first occurrence in the list of all of the specified
        values.
        """

        propVals = self.getPropertyValueList()
        listVals = [pv.get() for pv in propVals]

        for v in values:
            propVals.pop(listVals.index(v))

        PropertyValue.set(self, propVals)


    def reorder(self, idxs):
        """Reorders the list according to the given sequence of indices."""

        idxs = list(idxs)

        if list(sorted(idxs)) != list(range(len(self))):
            raise ValueError('Indices ({}) must '
                             'cover the list range '
                             '([0..{}])'.format(idxs, len(self) - 1))

        if idxs == list(range(len(self))):
            return

        propVals = self.getPropertyValueList()
        propVals = [propVals[i] for i in idxs]

        PropertyValue.set(self, propVals)


    def __setitem__(self, key, values):
        """Sets the value(s) of the list at the specified index/slice."""

        if isinstance(key, slice):
            indices = list(range(*key.indices(len(self))))
            if len(indices) != len(self) and \
               len(indices) != len(values):
                raise IndexError(
                    'PropertyValueList does not support complex slices')

        elif isinstance(key, int):
            indices = [key]
            values  = [values]
        else:
            raise IndexError('Invalid key type')

        # Replacement of all items in list
        if len(indices) == len(self) and \
           len(indices) != len(values):

            notifState = self.getNotificationState()
            self.disableNotification()
            del self[:]
            self.setNotificationState(notifState)
            self.extend(values)
            return

        # prepare the new values
        propVals    = self.getPropertyValueList()
        oldVals     = [pv.get() for pv in propVals]
        changedVals = [False] * len(self)

        # Update the PV instances that
        # correspond to the new values,
        # but suppress notification on them
        for idx, val in zip(indices, values):

            propVal    = propVals[idx]
            notifState = propVal.getNotificationState()

            propVal.disableNotification()
            propVal.set(val)
            propVal.setNotificationState(notifState)

            changedVals[idx] = not self._itemEqualityFunc(
                propVal.get(), oldVals[idx])

        # Notify list-level and item-level listeners
        # if any values in the list were changed
        if any(changedVals):

            log.debug('Notifying list-level listeners ({}.{} {})'.format(
                self._context().__class__.__name__,
                self._name,
                id(self._context())))
            self.propNotify()

            log.debug('Notifying item-level listeners ({}.{} {})'.format(
                self._context().__class__.__name__,
                self._name,
                id(self._context())))

            # The PV items will inevitably call
            # _listPVChanged, which will trigger
            # list-level notification - we just
            # performed list-level notification
            # above, so we want this suppressed.
            notifState = self.getNotificationState()
            self.disableNotification()

            for idx in indices:
                if changedVals[idx]:
                    propVals[idx].propNotify()

            self.setNotificationState(notifState)


    def __delitem__(self, key):
        """Remove items at the specified index/slice from the list."""

        propVals = self.getPropertyValueList()
        propVals.__delitem__(key)
        PropertyValue.set(self, propVals)


def safeCall(func, *args, **kwargs):
    """This function is may be used to "safely" run a function which may
    trigger ``PropertyValue`` notifications. Any notifications are queued
    and executed in the correct order.
    """
    name = uuid.uuid4()
    PropertyValue.queue.call(func, name, *args, **kwargs)



#!/usr/bin/env python
#
# bindable.py - This module adds functionality to the HasProperties class
# to allow properties from different instances to be bound to each other.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module adds functionality to the :class:`.HasProperties` class to
allow properties from different instances to be bound to each other.  The
logic defined in this module is separated purely to keep the
:mod:`.properties` and :mod:`.properties_value` module file sizes down.


The following functions are defined in this module and are used by the
the :class:`.HasProperties` class.


 .. autosummary::
    :nosignatures:

    bindProps
    unbindProps
    isBound


These functions use the following functions, which work directly with
:class:`.PropertyValue` instances, and are available for advanced usage:


 .. autosummary::
    :nosignatures:

    bindPropVals
    propValsAreBound


:class:`.PropertyValue` instances use the following methods for
synchronisation of attribute and values, notification of their listeners, and
access to other ``PropertyValue`` instances to which they are bound:

 .. autosummary::
    :nosignatures:

    syncAndNotify
    syncAndNotifyAtts
    buildBPVList


-------------
Example usage
-------------

::

    >>> import fsleyes_props as props

    >>> class MyObj(props.HasProperties):
            myint  = props.Int()
            myreal = props.Real()

    >>> myobj1 = MyObj()
    >>> myobj2 = MyObj()

    # Set some initial values
    >>> myobj1.myint  = 1
    >>> myobj1.myreal = 0.1
    >>> myobj2.myint  = 2
    >>> myobj2.myreal = 0.12

    # Bind myobj1.myint and myobj2.myint.
    # The instance on which bindProps is
    # called will inherit the value of the
    # other instance
    >>> myobj2.bindProps('myint', myobj1)

    >>> print myobj2
    MyObj
       myint = 1
      myreal = 0.2

    # Changing a bound property value on either
    # instance will result in the value being
    # propagated to the other instance.
    >>> myobj2.myint = 8
    >>> print myobj1
    MyObj
       myint = 8
      myreal = 0.1


-------
Details
-------


When a ``HasProperties`` property value is changed, the associated
``PropertyValue`` instance does two things:

  1. Casts and validates the new value and updates its stored value.
  2. Calls the :func:`syncAndNotify` function.

The ``syncAndNotify`` function then does the following:

  3. Updates the value on all bound ``PropertyValue`` instances.
  4. Notifies all listeners, registered on the source ``PV`` instance, of the
     value change.
  5. Notifies all listeners, registered on the bound ``PV`` instances, of the
     value change.

An important point to note regarding step 2 above is that *all* PV instances
which are bound, either directly or indirectly, to the source PV instance,
will be updated. In other words, there are no restrictions on the ways in
which ``PropertyValue`` instances may be bound.  A tree, chain, or even a
network of ``PV`` instances can be bound together - the above process will
still work.
"""


import logging
import weakref

import fsl.utils.weakfuncref as weakfuncref


log = logging.getLogger(__name__)


class Bidict(object):
    """A bare-bones bi-directional dictionary, used for binding
    :class:`.PropertyValueList` instances - see the :func:`_bindListProps` and
    :func:`_syncPropValLists` functions.
    """

    def __init__(self):
        self._thedict = {}

    def __setitem__(self, key, value):
        self._thedict[key]   = value
        self._thedict[value] = key

    def __delitem__(self, key):
        val = self._thedict.pop(key)
        self      ._thedict.pop(val)

    def get(self, key, default=None):
        return self._thedict.get(key, default)

    def __getitem__(self, key):
        return self._thedict.__getitem__(key)
    def __repr__(   self):
        return self._thedict.__repr__()
    def __str__(    self):
        return self._thedict.__str__()


def bindProps(self,
              propName,
              other,
              otherPropName=None,
              bindval=True,
              bindatt=True,
              unbind=False):
    """Binds the properties specified by ``propName`` and ``otherPropName``
    such that changes to one are applied to the other.

    If the properties are :class:`.List` properties, the :func:`_bindListProps`
    function is called. Otherwise the :func:`_bindProps` function is called.

    :arg str propName:  The name of a property on this ``HasProperties``
                        instance.

    :arg other:         Another ``HasProperties`` instance.

    :arg otherPropName: The name of a property on ``other`` to
                        bind to. If ``None`` it is assumed that
                        there is a property on ``other`` called
                        ``propName``.

    :arg bindval:       If ``True`` (the default), property values
                        are bound. This parameter is ignored for
                        list properties.

    :arg bindatt:       If ``True`` (the default), property attributes
                        are bound.  For :class:`.List` properties, this
                        parameter applies to the list values, not to the
                        list itself.

    :arg unbind:        If ``True``, the properties are unbound.
                        See the :meth:`unbindProps` method.
    """

    from . import properties

    if otherPropName is None: otherPropName = propName

    myProp    = self .getProp(propName)
    otherProp = other.getProp(otherPropName)

    if type(myProp) != type(otherProp):
        raise ValueError('Properties must be of the '
                         'same type to be bound')

    if isinstance(myProp, properties.ListPropertyBase):
        _bindListProps(self,
                       myProp,
                       other,
                       otherProp,
                       bindatt=bindatt,
                       unbind=unbind)
    else:
        _bindProps(self,
                   myProp,
                   other,
                   otherProp,
                   bindval=bindval,
                   bindatt=bindatt,
                   unbind=unbind)


def unbindProps(self,
                propName,
                other,
                otherPropName=None,
                bindval=True,
                bindatt=True):
    """Unbinds two properties previously bound via a call to :func:`bindProps`.
    """
    self.bindProps(propName,
                   other,
                   otherPropName,
                   bindval=bindval,
                   bindatt=bindatt,
                   unbind=True)


def isBound(self, propName, other, otherPropName=None):
    """Returns ``True`` if the specified property is bound to the
    other ``HasProperties`` instance, ``False`` otherwise.
    """

    if otherPropName is None: otherPropName = propName

    myProp       = self     .getProp(   propName)
    otherProp    = other    .getProp(   otherPropName)
    myPropVal    = myProp   .getPropVal(self)
    otherPropVal = otherProp.getPropVal(other)

    return propValsAreBound(myPropVal, otherPropVal)


def syncAndNotifyAtts(self, name, value):
    """This method is called by the
    :meth:`.PropertyValue.notifyAttributeListeners` method.

    It ensures that the attributes of any bound :class:`.PropertyValue`
    instances are synchronised, and then notifies all attribute listeners.
    """

    boundPropVals = _sync(self, True, name, value)
    boundPropVals = [self] + [bpv[0] for bpv in boundPropVals]

    _callAllListeners(boundPropVals, True, name, value)


def syncAndNotify(self):
    """Synchronises the value contained in all bound :class:`.PropertyValue`
    instances with the value contained in this instance, and notifies all
    registered listeners of the value change. This method is called by the
    :meth:`.PropertyValue.propNotify` method.
    """

    from . import properties_value

    bpvs    = _sync(self)
    allBpvs = []

    for bpv, listItems in bpvs:
        if isinstance(bpv, properties_value.PropertyValueList):
            allBpvs.extend(listItems)
        allBpvs.append(bpv)

    _callAllListeners([self] + allBpvs, False)


def _bindProps(self,
               myProp,
               other,
               otherProp,
               bindval=True,
               bindatt=True,
               unbind=False):
    """Binds the :class:`.PropertyValue` instances of two
    :class:`.PropertyBase` instances together. See the :func:`bindProps`
    function for details on the parameters.
    """

    myPropVal    = myProp   .getPropVal(self)
    otherPropVal = otherProp.getPropVal(other)

    if not unbind:
        allow = myPropVal.allowInvalid()
        myPropVal.allowInvalid(True)

        if bindatt: myPropVal.setAttributes(otherPropVal.getAttributes())
        if bindval: myPropVal.set(          otherPropVal.get())

        myPropVal.allowInvalid(allow)

    bindPropVals(myPropVal,
                 otherPropVal,
                 bindval=bindval,
                 bindatt=bindatt,
                 unbind=unbind)


def _bindListProps(self,
                   myProp,
                   other,
                   otherProp,
                   bindatt=True,
                   unbind=False):
    """Binds the :class:`.PropertyValueList` instances of two
    :class:`.ListPropertyBase` instances together. See the :func:`bindProps`
    function for details on the parameters.
    """

    myPropVal    = myProp   .getPropVal(self)
    otherPropVal = otherProp.getPropVal(other)

    # The unbinding case is easy
    if unbind:
        myPropValList    = myPropVal   .getPropertyValueList()
        otherPropValList = otherPropVal.getPropertyValueList()

        myPropVal   ._listPropValMaps.pop(id(otherPropVal))
        otherPropVal._listPropValMaps.pop(id(myPropVal))

        for myItem, otherItem in zip(myPropValList, otherPropValList):
            bindPropVals(myItem,
                         otherItem,
                         bindval=False,
                         bindatt=bindatt,
                         unbind=True)

        bindPropVals(myPropVal, otherPropVal, unbind=True)
        return

    # Binding two lists is a
    # bit more complicated ...

    # Inhibit list-level notification due to item
    # changes during the initial sync - we'll
    # manually do a list-level notification after
    # all the list values have been synced
    notifState = myPropVal.getNotificationState()
    myPropVal.disableNotification()

    # Force the two lists to have
    # the same number of elements
    if len(myPropVal) > len(otherPropVal):
        del myPropVal[len(otherPropVal):]

    elif len(myPropVal) < len(otherPropVal):
        myPropVal.extend(otherPropVal[len(myPropVal):])

    # Create a mapping between the
    # PropertyValue pairs across
    # the two lists
    myPropValList    = myPropVal   .getPropertyValueList()
    otherPropValList = otherPropVal.getPropertyValueList()
    propValMap       = Bidict()

    # Copy item values from the master list
    # to the slave list, and save the mapping
    for myItem, otherItem in zip(myPropValList, otherPropValList):

        log.debug('Binding list item {}.{} ({}) <- {}.{} ({})'.format(
            self.__class__.__name__,
            myProp.getLabel(self),
            myItem.get(),
            other.__class__.__name__,
            otherProp.getLabel(other),
            otherItem.get()))

        # Disable item notification - we'll
        # manually force a notify after the
        # sync
        itemNotifState = myItem.getNotificationState()
        myItem.disableNotification()

        # Bind attributes between PV item pairs,
        # but not value - value change of items
        # in a list is handled at the list level
        bindPropVals(myItem, otherItem, bindval=False, bindatt=bindatt)
        propValMap[myItem] = otherItem

        atts = otherItem.getAttributes()

        # Set attributes first, because the attribute
        # values may influence/modify the property value
        if bindatt: myItem.setAttributes(atts)
        myItem.set(otherItem.get())

        # Notify item level listeners of the value
        # change (if notification was enabled).
        #
        # TODO This notification occurs even
        # if the two PV objects had the same
        # value before the sync - you should
        # notify only if the myItem PV value
        # has changed.
        myItem.setNotificationState(itemNotifState)
        if itemNotifState:
            # notify attribute listeners first
            if bindatt:
                for name, val in atts.items():
                    syncAndNotifyAtts(myItem, name, val)

            syncAndNotify(myItem)

    # This mapping is stored on the PVL objects,
    # and used by the _syncListPropVals function
    myPropValMaps    = getattr(myPropVal,    '_listPropValMaps', {})
    otherPropValMaps = getattr(otherPropVal, '_listPropValMaps', {})

    # We can't use the PropValList objects as
    # keys, because they are not hashable.
    myPropValMaps[   id(otherPropVal)] = propValMap
    otherPropValMaps[id(myPropVal)]    = propValMap

    myPropVal   ._listPropValMaps = myPropValMaps
    otherPropVal._listPropValMaps = otherPropValMaps

    # Bind list-level value/attributes
    # between the PropertyValueList objects
    atts = otherPropVal.getAttributes()
    myPropVal.setAttributes(atts)

    bindPropVals(myPropVal, otherPropVal)

    # Manually notify list-level listeners
    #
    # TODO This notification will occur
    # even if the two lists had the same
    # value before being bound. It might
    # be worth only performing the
    # notification if the list has changed
    # value
    myPropVal.setNotificationState(notifState)

    # Sync the PVS, ensure that the sync
    # is propagated to other bound PVs,
    # and notify all listeners.
    for name, val in atts.items():
        syncAndNotifyAtts(myPropVal, name, val)

    syncAndNotify(myPropVal)


def propValsAreBound(pv1, pv2):
    """Returns ``True`` if the given :class:`.PropertyValue` instances are
    bound to each other, ``False`` otherwise.
    """

    pv1BoundPropVals = pv1.__dict__.get('boundPropVals', {})
    pv2BoundPropVals = pv2.__dict__.get('boundPropVals', {})

    return (id(pv2) in pv1BoundPropVals and
            id(pv1) in pv2BoundPropVals)


def bindPropVals(myPropVal,
                 otherPropVal,
                 bindval=True,
                 bindatt=True,
                 unbind=False):
    """Binds two :class:`.PropertyValue` instances together such that when the
    value of one changes, the other is changed. Note that the values are not
    immediately synchronised - they will become synchronised on the next change
    to either ``PropertyValue``.

    See :func:`bindProps` for details on the parameters.
    """

    mine  = myPropVal
    other = otherPropVal

    # A dict containing { id(PV) : PV } mappings is stored
    # on each PV, and used to maintain references to bound
    # PVs. We use a WeakValueDictionary (instead of just a
    # set) so that these references do not prevent PVs
    # which are no longer in use from being GC'd.
    wvd = weakref.WeakValueDictionary

    myBoundPropVals       = mine .__dict__.get('boundPropVals',    wvd())
    myBoundAttPropVals    = mine .__dict__.get('boundAttPropVals', wvd())
    otherBoundPropVals    = other.__dict__.get('boundPropVals',    wvd())
    otherBoundAttPropVals = other.__dict__.get('boundAttPropVals', wvd())

    if unbind: action = 'Unbinding'
    else:      action = 'Binding'

    log.debug('{} property values '
              '(val={}, att={}) {}.{} ({}) <-> {}.{} ({})'.format(
                  action,
                  bindval,
                  bindatt,
                  myPropVal._context.__class__.__name__,
                  myPropVal._name,
                  id(myPropVal),
                  otherPropVal._context.__class__.__name__,
                  otherPropVal._name,
                  id(otherPropVal)))

    if bindval:
        if unbind:
            myBoundPropVals   .pop(id(other))
            otherBoundPropVals.pop(id(mine))
        else:
            myBoundPropVals[   id(other)] = other
            otherBoundPropVals[id(mine)]  = mine

    if bindatt:
        if unbind:
            myBoundAttPropVals   .pop(id(other))
            otherBoundAttPropVals.pop(id(mine))
        else:
            myBoundAttPropVals[   id(other)] = other
            otherBoundAttPropVals[id(mine)]  = mine

    mine .boundPropVals    = myBoundPropVals
    mine .boundAttPropVals = myBoundAttPropVals
    other.boundPropVals    = otherBoundPropVals
    other.boundAttPropVals = otherBoundAttPropVals

    # When a master PV is synchronised to a slave PV,
    # it stores a flag on the slave PV which is checked
    # before starting a sync. If the flag is True,
    # the sync is inhibited. See the _sync function below.
    mine ._syncing = getattr(mine,  '_syncing', False)
    other._syncing = getattr(other, '_syncing', False)


def _syncPropValLists(masterList, slaveList):
    """Called by the :func:`_sync` function when one of a pair of bound
    :class:`.PropertyValueList` instances changes.

    Propagates the change on the ``masterList`` (either an addition, a
    removal, or a re-ordering) to the ``slaveList``.
    """

    propValMap = masterList._listPropValMaps[id(slaveList)]

    # If the change was due to the values of one or more PV
    # items changing (as opposed to a list modification -
    # addition/removal/reorder), the PV objects which
    # changed are stored in this list and returned
    changed = []

    # one or more items have been
    # added to the master list
    if len(masterList) > len(slaveList):

        # Loop through the PV objects in the master
        # list, and search for any which do not have
        # a paired PV object in the slave list
        for i, mpv in enumerate(masterList.getPropertyValueList()):

            spv = propValMap.get(mpv, None)

            # we've found a value in the master
            # list which is not in the slave list
            if spv is None:

                # add a new value to the slave list
                slaveList.insert(i, mpv.get())

                # retrieve the corresponding PV
                # object that was created by
                # the slave list
                spvs = slaveList.getPropertyValueList()
                spv  = spvs[i]

                # register a mapping between the
                # new master and slave PV objects
                propValMap[mpv] = spv

                # Bind the attributes of
                # the two new PV objects
                bindPropVals(mpv, spv, bindval=False)

    # one or more items have been
    # removed from the master list
    elif len(masterList) < len(slaveList):

        mpvs = masterList.getPropertyValueList()

        # Loop through the PV objects in the slave
        # list, and check to see if their mapped
        # master PV object has been removed from
        # the master list. Loop backwards so we can
        # delete items from the slave list as we go,
        # without having to offset the list index.
        for i, spv in reversed(
                list(enumerate(slaveList.getPropertyValueList()))):

            # If this raises an error, there's a bug
            # in somebody's code ... probably mine.
            mpv = propValMap[spv]

            # we've found a value in the slave list
            # which is no longer in the master list
            if mpv not in mpvs:

                # Delete the item from the slave
                # list, and delete the PV mapping
                del slaveList[ i]
                del propValMap[mpv]

    # list re-order, or individual
    # value change
    else:

        mpvs     = masterList.getPropertyValueList()
        mpvids   = [id(m) for m in mpvs]
        newOrder = []

        # loop through the PV objects in the slave list,
        # and build a list of indices of the corresponding
        # PV objects in the master list
        for i, spv in enumerate(slaveList.getPropertyValueList()):

            mpv = propValMap[spv]
            newOrder.append(mpvids.index(id(mpv)))

        # If the master list order has been
        # changed, re-order the slave list
        if newOrder != list(range(len(slaveList))):
            slaveList.reorder(newOrder)

        # The list order hasn't changed, so
        # this call must have been triggered
        # by a value change. Find the items
        # which have changed, and copy the
        # new value across to the slave list
        else:

            for i, (masterVal, slaveVal) in \
                enumerate(
                    zip(masterList.getPropertyValueList(),
                        slaveList .getPropertyValueList())):

                if masterVal == slaveVal: continue

                notifState = slaveVal.getNotificationState()
                validState = slaveVal.allowInvalid()
                slaveVal.disableNotification()
                slaveVal.allowInvalid(True)

                log.debug('Syncing bound PV list item '
                          '[{}] {}.{}({}) -> {}.{}({})'.format(
                              i,
                              masterList._context.__class__.__name__,
                              masterList._name,
                              masterVal.get(),
                              slaveList._context.__class__.__name__,
                              slaveList._name,
                              slaveList.get()))

                slaveVal.set(masterVal.get())
                changed.append(slaveVal)

                slaveVal.allowInvalid(validState)
                slaveVal.setNotificationState(notifState)

    return changed


def buildBPVList(self, key, node=None, bpvSet=None):
    """Recursively builds a list of all PVs that are bound to this one, either
    directly or indirectly.

    For each PV, we also store a reference to the 'parent' PV, i.e. the PV to
    which it is directly bound, as the direct bindings are needed to
    synchronise list PVs.

    Returns two lists - the first containing bound PVs, and the second
    containing the parent for each bound PV.

    :arg self:   The root PV.

    :arg key:    A string, either ``boundPropVals`` or ``boundAttPropVals``.

    :arg node:   The current PV to begin this step of the recursive search
                 from (do not pass in on the non-recursive call).

    :arg bpvSet: A set used to prevent cycles in the depth-first search (do
                 not pass in on the non-recursive call).
    """

    boundPropVals = []
    bpvParents    = []

    if node is None:
        node = self

    # A recursive depth-first search from this
    # PV through the network of all directly
    # or indirectly bound PVs.
    #
    # We use a set of PV ids to make sure
    # that we don't add duplicates to the
    # list of PVs that need to be synced
    if bpvSet is None:
        bpvSet = set()

    bpvs = node.__dict__.get(key, {}).values()
    bpvs = [b for b in bpvs if b is not self and id(b) not in bpvSet]

    for b in bpvs:
        bpvSet.add(id(b))

    boundPropVals.extend(bpvs)
    bpvParents   .extend([node] * len(bpvs))

    for bpv in bpvs:
        childBpvs, childBpvps = buildBPVList(self, key, bpv, bpvSet)

        boundPropVals.extend(childBpvs)
        bpvParents   .extend(childBpvps)

    return boundPropVals, bpvParents


def _sync(self, atts=False, attName=None, attValue=None):
    """Called by :func:`_notify`.

    Synchronises the value or attributes of all bound ``PropertyValue``
    instances to the the value or attributes of this one.

    :arg atts:     If ``True``, the attribute with ``attName`` and ``attValue``
                   is synchronised. Otherwise, the property values are
                   synchronised.

    :arg attName:  If ``att=True``, the name of the attribute to synchronise.

    :arg attValue: If ``att=True``, the value of the attribute to synchronise.
    """

    from . import properties_value

    # This PV is already being synced
    # to some other PV - don't sync back
    if getattr(self, '_syncing', False):
        return []

    if atts: key = 'boundAttPropVals'
    else:    key = 'boundPropVals'

    boundPropVals, bpvParents = buildBPVList(self, key)

    # Sync all the values that need syncing. Store
    # a ref to each PV which was synced, but not
    # to PVs which already had the same value.
    changedPropVals = []
    for i, bpv in enumerate(boundPropVals):

        # Don't bother if the values are already equal
        if atts:
            try:
                if bpv.getAttribute(attName) == attValue: continue
            except KeyError:
                pass
        elif self == bpv:
            continue

        # Set the syncing flag to prevent
        # recursive syncs back to this PV
        bpv._syncing = True

        # Disable notification on the PV, as we
        # manually trigger notifications in the
        # _notify function below.
        notifState = bpv.getNotificationState()
        bpv.disableNotification()

        log.debug('Syncing bound property values ({}) '
                  '{}.{} ({}) - {}.{} ({})'.format(
                      'attributes' if atts else 'values',
                      self._context.__class__.__name__,
                      self._name,
                      id(self._context),
                      bpv._context.__class__.__name__,
                      bpv._name,
                      id(bpv._context)))

        # Normal PropertyValue object (i.e. not a PropertyValueList)
        if atts or not isinstance(self, properties_value.PropertyValueList):

            # Store a reference to this PV
            changedPropVals.append((bpv, None))

            # Allow invalid values, as otherwise
            # an error may be raised.
            validState = bpv.allowInvalid()
            bpv.allowInvalid(True)

            # Sync the attribute value
            if atts: bpv.setAttribute(attName, attValue)

            # Or sync the property value
            else:    bpv.set(self.get())

            bpv.allowInvalid(validState)

        # PropertyValueList instances -
        # store a reference ot the PV list,
        # and to all list items that changed
        else:
            listItems = _syncPropValLists(bpvParents[i], bpv)
            changedPropVals.append((bpv, listItems))

        # Restore the notification state,
        # and remove the syncing flag
        bpv.setNotificationState(notifState)
        bpv._syncing = False

    # Return a list of all changed PVs back
    # to the _notify function, so it can
    # trigger notification on all of them.
    return changedPropVals


def _callAllListeners(propVals, att, name=None, value=None):
    """Calls all listeners of the given list of :class:`.PropertyValue`
    instances.

    :arg att:   If ``True``, attribute listeners are notified, otherwise
                value listeners are notified.
    :arg name:  If ``att == True``, the attribute name.
    :arg value: If ``att == True``, the attribute value.
    """

    from . import properties_value

    queued = []
    q      = properties_value.PropertyValue.queue

    # Hold the queue to inhibit any callbacks which
    # are triggered by immediate listener calls
    q.hold()

    # Get the function from a
    # properties_value.Listener
    # instance.
    def getFunc(listener):
        func = listener.function

        if isinstance(func, weakfuncref.WeakFunctionRef):
            func = func.function()

        return func

    try:
        for i, pv in enumerate(propVals):

            cListeners, cArgs = pv.prepareListeners(att, name, value)

            # If a bound PV is an item in a PV list,
            # then the listeners on the owning PV list
            # need to be called as well. We skip
            # the first PV in the propVals list, as it
            # is the source of the change.
            if (i > 0) and (not att) and (pv.getParent() is not None):
                pListeners, pArgs = pv.getParent().prepareListeners(False)
            else:
                pListeners = []
                pArgs      = None

            for listeners, args in [(cListeners, cArgs), (pListeners, pArgs)]:

                for l in listeners:

                    # The listener may have been removed/disabled
                    # due to another immediate listener
                    if not l.enabled:
                        continue

                    # Call the listener function directly
                    if l.immediate:
                        getFunc(l)(*args)

                    # Or add it to the queue
                    else:
                        queued.append((l, args))

    # Make sure the queue is freed
    finally:
        q.release()

    # Some listeners may have been disabled/removed
    # as the result of the execution of another
    # listener, so we only want to re-queue the ones
    # that are still active.
    queued = [(getFunc(l), l.makeQueueName(), a, {})
              for l, a in queued
              if l.enabled]

    # Some listeners referred to by weakrefs
    # may have been GC-d, in which case the
    # function reference will be None
    queued = [(f, n, a, kw) for f, n, a, kw in queued if f is not None]

    # Append any held functions on to the
    # end of the call list, so they are
    # executed after all of the listeners
    # for the original property value change
    held = q.clearHeld()
    q.callAll(queued + held)



#!/usr/bin/env python
#
# build_parts.py - Parts used by the build module to build a GUI from a
# HasProperties object.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""Parts for building a GUI.

This module provides definitiions of the parts used by the :mod:`.build`
module to build a GUI for a :class:`.HasProperties` instance. See the
:mod:`.build` module documentation for some examples.


----------------------------
Conditional visibility/state
----------------------------


The value of the ``ViewItem.dependencies`` parameter determines the required
format of the ``enabledWhen`` and ``visibleWhen`` parameters, and the
behaviour which controls the visible/enabled state of the displayed item.


If ``dependencies`` is ``None``, the visible/enabled state of the item will be
evaluated whenever any property on the target ``HasProperties`` instance
change. The ``enabledWhen`` and ``visibleWhen`` functions will be passed this
instance.


If ``dependencies`` is not ``None``, it must be a list containing the names of
properties on the ``HasProperties`` instance. The visible/enabled state of the
item will then be evaluated whenever the value of these specific properties
change, rather than whenever the value of any property changes. In this case,
the ``enabledWhen`` and ``visibleWhen`` functions will be passed the instance,
and the values of the specified properties as positional arguments.


If the enabled/visible state of the item is dependent on the value of a
property of a different ``HasProperties`` instance, the ``dependencies`` list
can contain a tuple which specifies the dependency as an ``(instance,
propName)`` pair. The ``instance`` may alternately be a function which accepts
the primary ``HasProperties`` instance as a single argument, and returns the
secondary instance (this is useful if you are defining ``ViewItems`` before
any of the ``HasProperties`` instances exist).


This all sounds a bit convoluted, but in practice is pretty simple.  Example::


    import wx
    import fsleyes_props as props

    class MyObj1(props.HasProperties):
        myPropA = props.Boolean()

    class MyObj2(props.HasProperties):
        myProp1 = props.Int()
        myProp2 = props.Real()
        myProp3 = props.Boolean()

        def __init__(self, mo1):
            self.mo1 = mo1

    mo2View = props.VGroup((

        # myProp1 is only visible
        # when MyObj1.myPropA is True
        props.Widget(
            'myProp1',

            # MyObj2 stores a reference to the
            # MyObj1 instance which controls
            # the state of the myProp1 widget.
            dependencies=[(lambda i: i.mo1, 'myPropA')],

            # The visibleWhen callback gets
            # passed the MyObj1 instance,
            # and the value of MyObj2.mo1.myPropA
            visibleWhen=lambda i, mo1val: mo1val),

        # myProp2 is only enabled
        # when MyObj2.myProp3 is True
        props.Widget(
            'myProp2',
            enabledWhen=lambda i: i.myProp3),

        # myProp3 is alway visible/enabled
        'myProp3'))

    mo1 = MyObj1()
    mo2 = MyObj2(mo1)

    app      = wx.App()
    mo1Frame = props.buildGUI(None, mo1)
    mo2Frame = props.buildGUI(None, mo2, view=mo2View)

    mo1Frame.Fit()
    mo2Frame.Fit()
    mo1Frame.Show()
    mo2Frame.Show()
    app.MainLoop()
"""


import copy

import logging


log = logging.getLogger(__name__)


class ViewItem(object):
    """Superclass for :class:`Widget`, :class:`Button`, :class:`Label` and
    :class:`Group`. Represents an item to be displayed.
    """

    def __init__(self,
                 key=None,
                 label=None,
                 tooltip=None,
                 visibleWhen=None,
                 enabledWhen=None,
                 dependencies=None,
                 setup=None,
                 **kwargs):
        """Define a ``ViewItem``.

        :param str key:      An identifier for this item. If this item is a
                             :class:`Widget`, this should be the property
                             name that the widget edits. This key is used to
                             look up labels and tooltips, if they are passed
                             in as dicts (see :func:`.buildGUI`).

        :param str label:    A label for this item, which may be used in the
                             GUI.

        :param str tooltip:  A tooltip, which may be displayed when the user
                             hovers the mouse over the widget for this
                             :class:`ViewItem`.

        :param visibleWhen:  A function which takes at least one argument (see
                             note about the ``dependencies`` parameter above),
                             the ``HasProperties`` instance, and returns a
                             ``bool``. When any property values change, the
                             function is called. The return value is used to
                             determine whether this item should be made
                             visible or invisible.

        :param enabledWhen:  Same as the ``visibleWhen`` parameter, except the
                             state of the item (and its children) is changed
                             between enabled and disabled.

        :param dependencies: List of dependencies which determine when the
                             visible/enabled state of the item are evaluated,
                             and the arguments that are passed to the
                             ``visibleWhen``/``enabledWhen`` functions.

        :param setup:        Optional function which is called when te GUI
                             object represented by this ViewItem is first
                             created. The function is passed the
                             ``HasProperties`` instance, the :mod:`wx` GUI
                             parent object, and the :mod:`wx` object.

        :param kwargs:       Any type-specific options which are to be passed
                             through to the creation function, defined in
                             the :mod:`.widgets` module.
        """

        self.key          = key
        self.label        = label
        self.tooltip      = tooltip
        self.visibleWhen  = visibleWhen
        self.enabledWhen  = enabledWhen
        self.dependencies = dependencies
        self.setup        = setup
        self.kwargs       = kwargs


    def __str__(self):
        return '{}({})'.format(self.__class__.__name__, self.key)


    def __repr__(self):
        return self.__str__()


    def __deepcopy__(self, memo):
        """Creates a deep copy of this ``ViewItem``. Deep copies are made
        of all attributes, except for the ``dependencies`` list, which may
        contain references to objects that cannot be deep-copied.
        """

        new = self.__class__.__new__(self.__class__)

        memo[id(self)] = new

        for k, v in self.__dict__.items():
            if k == 'dependencies': setattr(new, k, copy.copy(    v))
            else:                   setattr(new, k, copy.deepcopy(v, memo))

        return new


class Button(ViewItem):
    """Represents a button which, when clicked, will call a specified callback
    function.

    When the button is clicked, the callback function is passed two arguemnts
    - the :class:`.HasProperties` instance, and the :class:`wx.Button`
    instance.
    """

    def __init__(self,
                 key=None,
                 text=None,
                 callback=None,
                 icon=None,
                 **kwargs):
        """
        If the ``icon`` argument is provided, it is used instead of any
        specified ``text``. The ``icon`` should be a string containing the
        name of the image file.

        :arg key:      Passed to :meth:`ViewItem.__init__`.
        :arg text:     Text to show on the button (if ``icons`` are not used).
        :arg callback: Function to call when the button is clicked.
        :arg icon:     The name of an icon image to show on the button.

        All other arguments are passed through to :meth:`ViewItem.__init__`.
        """
        self.callback = callback
        self.icon     = icon
        self.text     = text
        ViewItem.__init__(self, key, **kwargs)

    def __str__( self):
        return '{}({})'.format(self.__class__.__name__, self.text)
    def __repr__(self):
        return self.__str__()


class Toggle(ViewItem):
    """Represents a *toggle* widget of some sort (e.g. a check box, or a
    toggle button) which, when clicked, calls a specified callback function.
    """
    def __init__(self,
                 key=None,
                 callback=None,
                 icon=None,
                 **kwargs):
        """Create a ``Toggle``.

        :arg key:      Passed to :meth:`ViewItem.__init__`.
        :arg callback: Function to call when the toggle widget is clicked.
                       Must accept two arguments - the :class:`.HasProperties`
                       instance, and the toggle widget.
        :arg icon:     The name of an image file, or a list of two image
                       file names.

        All other arguments are passed through to :meth:`ViewItem.__init__`.
        """

        self.callback = callback
        self.icon     = icon
        ViewItem.__init__(self, key, **kwargs)


    def __str__( self):
        return '{}()'.format(self.__class__.__name__)
    def __repr__(self):
        return self.__str__()


class Label(ViewItem):
    """Represents a static text label."""


    def __init__(self, viewItem=None, **kwargs):
        """Define a label.

        :class:`Label` objects are automatically created for other
        :class:`ViewItem` objects, which are to be labelled.
        """

        if viewItem is not None:
            kwargs['key']          = '{}_label'.format(viewItem.key)
            kwargs['label']        = viewItem.label
            kwargs['tooltip']      = viewItem.tooltip
            kwargs['visibleWhen']  = viewItem.visibleWhen
            kwargs['enabledWhen']  = viewItem.enabledWhen
            kwargs['dependencies'] = viewItem.dependencies

        ViewItem.__init__(self, **kwargs)

    def __str__( self):
        return '{}({})'.format(self.__class__.__name__, self.label)
    def __repr__(self):
        return self.__str__()


class LinkBox(ViewItem):
    """Represents a checkbox which allows the user to control whether a
    property is linked (a.k.a. bound) to the parent of the
    :class:`HasProperties` object.
    """

    def __init__(self, viewItem=None, **kwargs):

        if viewItem is not None:
            self.propKey          = viewItem.key
            kwargs['key']         = '{}_linkBox'.format(viewItem.key)
            kwargs['label']       = viewItem.label
            kwargs['tooltip']     = viewItem.tooltip
            kwargs['visibleWhen'] = viewItem.visibleWhen
            kwargs['enabledWhen'] = viewItem.enabledWhen

        ViewItem.__init__(self, **kwargs)


class Widget(ViewItem):
    """Represents a widget which is used to modify a property value. """


    def __init__(self, propName, **kwargs):
        """Define a :class:`Widget`.

        :param str propName: The name of the property which this widget can
                             modify.

        :param kwargs:       Passed to the :class:`ViewItem` constructor.
        """

        kwargs['key'] = propName
        ViewItem.__init__(self, **kwargs)


class Group(ViewItem):
    """Represents a collection of other :class:`ViewItem` objects.

    This class is not to be used directly - use one of the subclasses:
      - :class:`VGroup`
      - :class:`HGroup`
      - :class:`NotebookGroup`
    """


    def __init__(self, children, showLabels=True, border=False, **kwargs):
        """Define a :class:`Group`.

        Parameters:

        :param children:        List of :class:`ViewItem` objects, the
                                children of this :class:`Group`.

        :param bool showLabels: Whether labels should be displayed for each of
                                the children. If this is ``True``, an attribute
                                will be added to this :class:`Group` object in
                                the :func:`_prepareView` function, called
                                ``childLabels``, which contains a
                                :class:`Label` object for each child.

        :param bool border:     If ``True``, this group will be drawn with a
                                border around it. If this group is a child of
                                another :class:`VGroup`, it will be laid out
                                a bit differently, too.

        :param kwargs:          Passed to the :class:`ViewItem` constructor.
        """
        ViewItem.__init__(self, **kwargs)
        self.children   = children
        self.border     = border
        self.showLabels = showLabels

    def __str__(self):

        def doStr(viewItem, depth):

            if isinstance(viewItem, Group):

                s = '{}{}\n'.format(' ' * depth, ViewItem.__str__(viewItem))
                for child in viewItem.children:
                    s = s + doStr(child, depth + 1)
            else:
                s = '{}{}\n'.format(' ' * depth, str(viewItem))
            return s

        return doStr(self, 0)


class NotebookGroup(Group):
    """A :class:`Group` representing a GUI Notebook. Children are added as
    notebook pages.
    """
    def __init__(self, children, **kwargs):
        """Define a :class:`NotebookGroup`.

        :param children: List of :class:`ViewItem` objects - a tab in the
                         notebook is added for each child.
        """
        Group.__init__(self, children, **kwargs)


class HGroup(Group):
    """A group representing a GUI panel, whose children are laid out
    horizontally.
    """

    def __init__(self, children, wrap=False, vertLabels=False, **kwargs):
        """Create a :class:`HGroup`.

        :arg wrap:       If ``True`` the children are wrapped, via a
                         :class:`wx.WrapSizer`; if there is not enough
                         horizontal space to display all children in a
                         single row, the remaining children are
                         displayed on a new row.

        :arg vertLabels: If ``True`` child labels are displayed above
                         the child.
        """
        self.wrap       = wrap
        self.vertLabels = vertLabels
        Group.__init__(self, children, **kwargs)


class VGroup(Group):
    """A group representing a GUI panel, whose children are laid out
    vertically.
    """
    def __init__(self, children, **kwargs):
        kwargs['border'] = kwargs.get('border', True)
        Group.__init__(self, children, **kwargs)



#!/usr/bin/env python
#
# syncable.py - An extension to the HasProperties class which allows
# a master-slave relationship to exist between instances.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`SyncableHasProperties` class, an extension
to the :class:`.HasProperties` class which allows a parent-child relationship
to exist between instances. A one-to-many synchronisation relationship is
possible between one parent, and many children. Property values are
synchronised between a parent and its children, using the functionality
provided by the :mod:`bindable` module.

All that is needed to make use of this functionality is to extend the
``SyncableHasProperties`` class instead of the ``HasProperties`` class::

    >>> import fsleyes_props as props

    >>> class MyObj(props.SyncableHasProperties):
            myint = props.Int()
            def __init__(self, parent=None):
                props.SyncableHasProperties.__init__(self, parent=parent)


Given a class definition such as the above, a parent-child relationship
between two instances can be set up as follows::

    >>> myParent = MyObj()
    >>> myChild  = MyObj(myParent)

The ``myint`` properties of both instances are now bound to each other - when
it changes in one instance, that change is propagated to the other instance::

    >>> def parentPropChanged(*a):
            print 'myParent.myint changed: {}'.format(myParent.myint)
    >>>
    >>> def childPropChanged(*a):
            print 'myChild.myint changed: {}'.format(myChild.myint)

    >>> myParent.addListener('myint', 'parentPropChanged', parentPropChanged)
    >>> myChild.addListener( 'myint', 'childPropChanged',  childPropChanged)

    >>> myParent.myint = 12345
    myParent.myint changed: 12345
    myChild.myint changed: 12345

    >>> myChild.myint = 54321
    myParent.myint changed: 54321
    myChild.myint changed: 54321

This synchronisation can be toggled on the child instance, via the
:meth:`unsyncFromParent` and :meth:`syncToParent` methods of the
:class:`SyncableHasProperties` class.  Listeners to sync state changes may
be registered on the child instance via the :meth:`addSyncChangeListener`
method (and de-registered via the :meth:`removeSyncChangeListener` method).
"""


import weakref
import logging

from . import properties       as props
from . import suppress         as suppress
from . import properties_types as types


log = logging.getLogger(__name__)


_SYNC_SALT_ = '_sync_'
"""Constant string added to sync-related property names and listeners."""


class SyncableHasProperties(props.HasProperties):
    """An extension to the ``HasProperties`` class which supports parent-child
    relationships between instances.
    """


    @classmethod
    def _saltSyncPropertyName(cls, propName):
        """Adds a prefix to the given property name, to be used as the
        name for the corresponding boolean sync property.
        """
        return '{}{}'.format(_SYNC_SALT_, propName)


    @classmethod
    def _unsaltSyncPropertyName(cls, propName):
        """Removes a prefix from the given property name, which was
        added by the :meth:`_saltSyncPropertyName` method.
        """
        return propName[len(_SYNC_SALT_):]


    @classmethod
    def getSyncPropertyName(cls, propName):
        """Returns the name of the boolean property which can be used to
        toggle binding of the given property to the parent property of
        this instance.
        """
        return cls._saltSyncPropertyName(propName)


    @classmethod
    def getSyncProperty(cls, propName):
        """Returns the :class:`.PropertyBase` instance of the boolean property
        which can be used to toggle binding of the given property to the parent
        property of this instance.
        """
        return cls.getProp(cls.getSyncPropertyName(propName))


    def __init__(self, **kwargs):
        """Create a ``SyncableHasProperties`` instance.

        If this ``SyncableHasProperties`` instance does not have a parent,
        there is no need to call this constructor explicitly. Otherwise, the
        parent must be an instance of the same class to which this instance's
        properties should be bound.

        :arg parent:    Another ``SyncableHasProperties`` instance, which has
                        the same type as this instance.

        :arg nobind:    A sequence of property names which should not be bound
                        with the parent.

        :arg nounbind:  A sequence of property names which cannot be unbound
                        from the parent.

        :arg state:     Initial synchronised state. Can be either ``True`` or
                        ``False``, in which case all properties will initially
                        be either synced or unsynced. Or can be a dictionary
                        of ``{propName : boolean}`` mappings, defining the sync
                        state for each property.

        :arg direction: Initial binding direction. Not applicable if this
                        instance does not have a parent. If ``True``, when a
                        property is bound to the parent, this instance will
                        inherit the parent's value. Otherwise, when a property
                        is bound, the parent will inherit this instance's
                        value.

        :arg kwargs:    Other arguments are passed to the
                        :meth:`.HasProperties.__init__` method.
        """

        parent    = kwargs.pop('parent',    None)
        nobind    = kwargs.pop('nobind',    [])
        nounbind  = kwargs.pop('nounbind',  [])
        state     = kwargs.pop('state',     True)
        direction = kwargs.pop('direction', True)

        props.HasProperties.__init__(self, **kwargs)

        self._nobind   = nobind
        self._nounbind = nounbind

        # If parent is none, then this instance
        # is a 'parent' instance, and doesn't need
        # to worry about being bound. So we've got
        # nothing to do.
        if parent is None:

            # This array maintains a list of
            # all the children synced to this
            # parent
            self._children = []
            self._parent   = None
            return

        # Otherwise, this instance is a 'child'
        # instance - make sure the parent is
        # valid (of the same type)
        if not isinstance(parent, self.__class__):
            raise TypeError('parent is of a different type '
                            '({} != {})'.format(parent.__class__,
                                                self.__class__))


        # Set up a binding between this
        # instance and its parent
        self._parent = weakref.ref(parent)
        parent._children.append(weakref.ref(self))

        # This dictionary contains
        #
        # { propName : boolean }
        #
        # mappings, indicating the binding
        # direction that should be used
        # when a property is synchronised
        # to the parent. A value of True
        # implies a parent -> child  binding
        # direction (i.e. the child will
        # inherit the value of the parent),
        # and a value of False implies a
        # child -> parent binding direction.
        self._bindDirections = {}

        log.debug('Binding properties of {} ({}) to parent ({})'.format(
            self.__class__.__name__, id(self), id(parent)))

        # Get a list of all the
        # properties of this class
        propNames, _ = self.getAllProperties()

        for pn in propNames:

            # Add a boolean sync property
            # for this regular property.
            bindProp = types.Boolean(default=True)

            self.addProperty(self._saltSyncPropertyName(pn), bindProp)

            # Initialise the binding direction
            # and initial state
            self._bindDirections[pn] = direction

            if isinstance(state, dict): pState = state.get(pn, True)
            else:                       pState = state

            if   not self.canBeSyncedToParent(    pn): pState = False
            elif not self.canBeUnsyncedFromParent(pn): pState = True

            self._initSyncProperty(pn, pState)


    def getParent(self):
        """Returns the parent of this instance, or ``None`` if there is no
        parent.

        On child ``SyncableHasProperties`` instances, this method must not
        be called before :meth:`__init__` has been called. If this happens,
        an :exc:`AttributeError` will be raised.
        """

        if self._parent is None: return None
        else:                    return self._parent()


    def getChildren(self):
        """Returns a list of all children that are synced to this parent
        instance, or ``None`` if this instance is not a parent.
        """
        if self._parent is not None:
            return None

        children = [c() for c in self._children]
        children = [c for c in children if c is not None]

        return children


    def _saltSyncListenerName(self, propName):
        """Adds a prefix and a suffix to the given property name, to be used
        as the name for an internal listener on the corresponding boolean sync
        property.

        """
        return '{}{}_{}'.format(_SYNC_SALT_, propName, id(self))


    def _initSyncProperty(self, propName, initState):
        """Called by child instances from :meth:`__init__`.

        Configures a binding between this instance and its parent for the
        specified property.
        """

        bindPropName  = self._saltSyncPropertyName(propName)
        bindPropObj   = self.getProp(bindPropName)
        bindPropVal   = bindPropObj.getPropVal(self)
        direction     = self._bindDirections[propName]

        if initState and not self.canBeSyncedToParent(propName):
            raise ValueError('Invalid initial state for '
                             'nobind property {}'.format(propName))

        if (not initState) and (not self.canBeUnsyncedFromParent(propName)):
            raise ValueError('Invalid initial state for '
                             'nounbindproperty {}'.format(propName))

        if not self.canBeSyncedToParent(propName):
            bindPropVal.set(False)
            return

        bindPropVal.set(initState)

        if self.canBeUnsyncedFromParent(propName):
            lName = self._saltSyncListenerName(propName)
            bindPropVal.addListener(lName, self._syncPropChanged)

        if initState:
            if direction: slave, master = self, self._parent()
            else:         slave, master = self._parent(), self
            slave.bindProps(propName, master)


    def _syncPropChanged(self, value, valid, ctx, bindPropName):
        """Called when a hidden boolean property controlling the sync
        state of the specified real property changes.

        Changes the sync state of the property accordingly.
        """

        propName    = self._unsaltSyncPropertyName(bindPropName)
        bindPropVal = getattr(self, bindPropName)
        direction   = self._bindDirections[propName]

        if bindPropVal and (propName in self._nobind):
            raise RuntimeError('{} cannot be bound to '
                               'parent'.format(propName))

        if (not bindPropVal) and (propName in self._nounbind):
            raise RuntimeError('{} cannot be unbound from '
                               'parent'.format(propName))

        log.debug('Sync property changed for {} - '
                  'changing binding state'.format(propName))

        if direction: slave, master = self, self._parent()
        else:         slave, master = self._parent(), self

        slave.bindProps(propName, master, unbind=(not bindPropVal))


    def getBindingDirection(self, propName):
        """Returns the current binding direction for the given property. See
        the :meth:`setBindingDirection` method.
        """
        return self._bindDirections[propName]


    def setBindingDirection(self, direction, propName=None):
        """Set the current binding direction for the named property. If the
        direction is ``True``, when this property is bound, this instance
        will inherit the parent's value. Otherwise, when this property
        is bound, the parent will inherit the value from this instance.

        If a property is not specified, the binding direction of all
        properties will be changed.
        """
        if propName is None: propNames = self._bindDirections.keys()
        else:                propNames = [propName]

        for pn in propNames:
            self._bindDirections[pn] = direction


    def syncToParent(self, propName):
        """Synchronise the given property with the parent instance.

        If this ``SyncableHasProperties`` instance has no parent, a
        :exc:`RuntimeError` is raised. If the specified property is in the
        ``nobind`` list (see :meth:`__init__`), a :exc:`RuntimeError` is
        raised.

        ..note:: The ``nobind`` check can be avoided by calling
        :func:`.bindable.bindProps` directly. But don't do that.
        """
        if propName in self._nobind:
            raise RuntimeError('{} cannot be bound to '
                               'parent'.format(propName))

        bindPropName = self._saltSyncPropertyName(propName)

        if getattr(self, bindPropName):
            return

        with suppress.suppress(self, bindPropName):
            setattr(self, bindPropName, True)

        self._syncPropChanged(None, None, None, bindPropName)


    def unsyncFromParent(self, propName):
        """Unsynchronise the given property from the parent instance.

        If this :class:`SyncableHasProperties` instance has no parent, a
        :exc:`RuntimeError` is raised. If the specified property is in the
        `nounbind` list (see :meth:`__init__`), a :exc:`RuntimeError` is
        raised.

        ..note:: The ``nounbind`` check can be avoided by calling
        :func:`bindable.bindProps` directly. But don't do that.
        """
        if propName in self._nounbind:
            raise RuntimeError('{} cannot be unbound from '
                               'parent'.format(propName))

        bindPropName = self._saltSyncPropertyName(propName)

        if not getattr(self, bindPropName):
            return

        with suppress.suppress(self, bindPropName):
            setattr(self, bindPropName, False)

        self._syncPropChanged(None, None, None, bindPropName)


    def syncAllToParent(self):
        """Synchronises all properties to the parent instance.

        Does not attempt to synchronise properties in the ``nobind`` list.
        """
        propNames = self.getAllProperties()[0]

        for propName in propNames:
            if propName in self._nounbind or \
               propName in self._nobind:
                continue

            self.syncToParent(propName)


    def unsyncAllFromParent(self):
        """Unynchronises all properties from the parent instance.

        Does not attempt to synchronise properties in the ``nounbind`` list.
        """
        propNames = self.getAllProperties()[0]

        for propName in propNames:
            if propName in self._nounbind or \
               propName in self._nobind:
                continue

            self.unsyncFromParent(propName)


    def detachFromParent(self):
        """If this is a child ``SyncableHasProperties`` instance, it
        detaches itself from its parent. This is an irreversible operation.

        TODO: Add the ability to dynamically set/clear the parent
              SHP instance.
        """

        # This is a parent instance -
        # nothing to detach from
        if self._parent is None:
            return

        parent    = self._parent()
        propNames = self.getAllProperties()[0]

        for propName in propNames:
            if propName not in self._nounbind:
                syncPropName = self._saltSyncPropertyName(propName)
                lName        = self._saltSyncListenerName(propName)

                self.removeListener(syncPropName, lName)

                # The parent may have been GC'd
                if parent is not None:
                    self.unsyncFromParent(propName)

        if parent is not None:
            for c in list(parent._children):
                if c() is self:
                    parent._children.remove(c)

        self._parent = None


    def isSyncedToParent(self, propName):
        """Returns ``True`` if the specified property is synced to the parent
        of this ``SyncableHasProperties`` instance, ``False`` otherwise.
        """
        return getattr(self, self._saltSyncPropertyName(propName))


    def anySyncedToParent(self):
        """Returns ``True`` if any properties are synced to the parent
        of this ``SyncableHasProperties`` instance, ``False`` otherwise.
        """
        propNames = self.getAllProperties()[0]
        return any([self.isSyncedToParent(p) for p in propNames])


    def allSyncedToParent(self):
        """Returns ``True`` if all properties are synced to the parent
        of this ``SyncableHasProperties`` instance, ``False`` otherwise.
        """
        propNames = self.getAllProperties()[0]
        return all([self.isSyncedToParent(p) for p in propNames])


    def canBeSyncedToParent(self, propName):
        """Returns ``True`` if the given property can be synced between this
        ``SyncableHasProperties`` instance and its parent (see the ``nobind``
        parameter in :meth:`__init__`).
        """
        return propName not in self._nobind


    def canBeUnsyncedFromParent(self, propName):
        """Returns ``True`` if the given property can be unsynced between this
        ``SyncableHasProperties`` instance and its parent (see the
        ``nounbind`` parameter in :meth:`__init__`).
        """
        return propName not in self._nounbind


    def addSyncChangeListener(self,
                              propName,
                              listenerName,
                              callback,
                              overwrite=False,
                              weak=True):
        """Registers the given callback function to be called when
        the sync state of the specified property changes.
        """
        bindPropName = self._saltSyncPropertyName(propName)
        self.addListener(bindPropName,
                         listenerName,
                         callback,
                         overwrite=overwrite,
                         weak=weak)


    def removeSyncChangeListener(self, propName, listenerName):
        """De-registers the given listener from receiving sync
        state changes.
        """
        bindPropName = self._saltSyncPropertyName(propName)
        self.removeListener(bindPropName, listenerName)
